# CircleIntersectTool
# Copyright 2016 SLBaumgartner
#
# March 8, 2016 Version 0.1 (Beta)
#
# This creates a pair of guide points at the intersection points of two circles
# selected by the user.  The points are calculated from the true mathematical
# representation of the circles, not from SketchUp's representation as a
# sequence of Edges, so they can be used for purposes such as triangulation.
# If the selections are partial arcs, the intersections are found for the
# full circles containing the arcs.

# This test version adds an item named "Circle Intersect" to the Tools
# menu in SketchUp.  It does not yet have a toolbar and is not yet packaged
# as an Extension.  To install it, you must copy this file into the Plugins
# folder.


module SLBPlugins
  module CircleIntersect

    class CircleIntersectTool

      def initialize
        @ip = Sketchup::InputPoint.new
        reset
      end

      def reset
        # state == 0 => waiting for first selection
        # state == 1 => waiting for second selection
        # state == 2 => ready to create cpoints
        @state = 0

        # arc curves selected by user
        @ac1 = nil
        @ac2 = nil

        Sketchup.active_model.selection.clear
      end

      def activate
        reset
        set_status
      end

      def set_status
        case @state
        when 0
          Sketchup.set_status_text 'select first arc'
        when 1
          Sketchup.set_status_text 'select second arc'
        end
      end

      def increment_state
        @state = @state + 1
        if @state == 2
          createCpoints
          reset
        end
        set_status
      end

    # need this to cause inference point and tooltips to be drawn
    def onMouseMove(flags, x, y, view)
      @ip.pick(view, x, y)
      view.invalidate
      nil
    end

    def onLButtonDown(flags, x, y, view)
        # see what the user has clicked
        status = @ip.pick(view, x, y)

        # InputPoint is flakey - when you pick an edge it will often
        # return the associated Vertex instead, so need some error
        # trapping here.
        edge = @ip.edge
        vertex = @ip.vertex
        if edge.nil? && ! vertex.nil?
          edges = vertex.edges
          edge = edges[0]
        end
        
        if edge.nil? || edge.curve.nil?
          UI.messagebox("Please select an arc")
          return
        else
          Sketchup.active_model.selection.add(edge.curve.edges)
          case @state
          when 0
            @ac1 = edge.curve
          when 1
            @ac2 = edge.curve
          end
          increment_state
        end
        nil
      end

      def createCpoints
        begin
          intersects = CICalcs.intersect(@ac1, @ac2)
        rescue Exception => e
          UI.messagebox e.message
          reset
          return
        end

        # intersections found, add cpoints to model
        ents=Sketchup.active_model.active_entities
        Sketchup.active_model.start_operation('CircleIntersect', true)
        ents.add_cpoint(intersects[0])
        ents.add_cpoint(intersects[1])
        Sketchup.active_model.commit_operation
        nil
      end
      
      
    end # CircleIntersectTool
    
    module CICalcs
      # Find the intersection points of two ArcCurves.  This
      # works based on the metadata about the real geometric
      # circles, not the SketchUp entity representation as a series
      # of Edge entities.  It generates true intersections, not
      # approximations based on the Edges.
      # This will work both for full circles and for partial arcs.
      # In the latter case, it will behave as if the arc is extended
      # into a full circle, ignoring the actual start and end.
      # TBD: should I change the last point to respect arcs vs circles?
      def self.intersect(c1, c2)
        unless c1.is_a?(Sketchup::ArcCurve) && c2.is_a?(Sketchup::ArcCurve)
          raise ArgumentError, 'arguments must be ArcCurves', caller
        end
        
        # test for coplanar by means of API methods rather than simply comparing
        # planes because the planes are subject to computation error and their
        # equality test does not have a tolerance.
        unless(c1.center.on_plane?(c2.plane) && c2.center.on_plane?(c1.plane))
          raise ArgumentError, 'Arcs are not coplanar', caller
        end

        center_vect = (c2.center - c1.center)
        d = center_vect.length
        # concentric arcs can never intersect.  Trapping d==0 here
        # avoids divide by zero later.
        if(d == 0.0)
          raise ArgumentError, 'Arcs have same center', caller
        end
        cross_vect = (c1.normal * center_vect).normalize
        center_vect.normalize!
        
        r1=c1.radius
        r2=c2.radius
        if(d > r1+r2)
          raise ArgumentError, 'Centers are too far apart', caller
        end
        
        x = (d**2 - r2**2 + r1**2)/(2*d)
        begin
          y = Math::sqrt(c1.radius**2 - x**2)
        rescue
          # I think the only case that causes this is if one circle
          # is inside the other, but handle it as generic no intersection
          # until I am absolutely sure
          raise ArgumentError, 'Arcs do not intersect', caller
        end

        center_vect.length = x
        cross_vect.length = y

        x_point = c1.center.offset(center_vect)

        # return the points of intersection.
        # If the circles just kiss, y==0 and the two points will be the same
        [x_point.offset(cross_vect), x_point.offset(cross_vect.reverse)]
      end # intersect
    end #CICalcs

    unless file_loaded?(__FILE__)
      cmd = UI::Command.new('Circle Intersect') {Sketchup.active_model.select_tool CircleIntersectTool.new}
      cmd.menu_text = 'Circle Intersect'
      UI.menu("Tools").add_item cmd
    end
    file_loaded(__FILE__)
  end # CircleIntersect
end # SLBPlugins
